# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

#
# pygeometa as a service
# ----------------------
#
# This file is intended to be used as a pygeoapi process plugin which will
# provide pygeometa functionality via OGC API - Processes.
#
# To integrate this plugin in pygeoapi:
#
# 1. ensure pygeometa is installed into the pygeoapi deployment environment
#
# 2. add the processes to the pygeoapi configuration as follows:
#
# pygeometa-metadata-schemas:
#     type: process
#     processor:
#         name: pygeometa.pygeoapi_plugin.PygeometaMetadataSchemasProcessor
#
# pygeometa-metadata-import:
#     type: process
#     processor:
#         name: pygeometa.pygeoapi_plugin.PygeometaMetadataImportProcessor
#
# pygeometa-metadata-validate:
#     type: process
#     processor:
#         name: pygeometa.pygeoapi_plugin.PygeometaMetadataValidateProcessor
#
# pygeometa-metadata-generate:
#     type: process
#     processor:
#         name: pygeometa.pygeoapi_plugin.PygeometaMetadataGenerateProcessor
#
# pygeometa-metadata-transform:
#     type: process
#     processor:
#         name: pygeometa.pygeoapi_plugin.PygeometaMetadataTransformProcessor
#
#
# 3. (re)start pygeoapi
#
# The resulting processes will be available at the following endpoints:
#
# /processes/pygeometa-metadata-schemas
#
# /processes/pygeometa-metadata-import
#
# /processes/pygeometa-metadata-validate
#
# /processes/pygeometa-metadata-generate
#
# /processes/pygeometa-metadata-transform
#
# Note that pygeoapi's OpenAPI/Swagger interface (at /openapi) will also
# provide a developer-friendly interface to test and run requests
#

import logging

from pygeometa import __version__
from pygeometa.core import read_mcf, validate_mcf
from pygeometa.schemas import get_supported_schemas, load_schema

from pygeoapi.process.base import BaseProcessor, ProcessorExecuteError


LOGGER = logging.getLogger(__name__)

INPUT_MCF = {
    'title': 'Metadata control file (MCF)',
    'description': 'a pygeometa metadata control file (MCF) as JSON',
    'schema': {
        'type': 'object',
        'contentMediaType': 'application/json'
    },
    'minOccurs': 1,
    'maxOccurs': 1,
    'metadata': None,
    'keywords': ['metadata control file', 'mcf']
}

INPUT_METADATA = {
    'title': 'Metadata',
    'description': 'Metadata record',
    'schema': {
        'type': 'string',
    },
    'minOccurs': 1,
    'maxOccurs': 1,
    'metadata': None,
    'keywords': ['metadata', 'record']
}

INPUT_SCHEMA = {
    'title': 'Metadata schema',
    'description': 'Metadata schema',
    'schema': {
        'type': 'string',
        'enum': list(get_supported_schemas())
    },
    'minOccurs': 1,
    'maxOccurs': 1,
    'metadata': None,
    'keywords': ['metadata', 'schema']
}


PROCESS_METADATA_SCHEMAS = {
    'version': __version__,
    'id': 'pygeometa-metadata-schemas',
    'title': {
        'en': 'pygeometa metadata schemas',
    },
    'description': {
        'en': 'List supported schemas'
    },
    'keywords': ['pygeometa', 'metadata', 'schemas'],
    'links': [{
        'type': 'text/html',
        'rel': 'about',
        'title': 'information',
        'href': 'https://geopython.github.io/pygeometa/pygeoapi-plugin',
        'hreflang': 'en-US'
    }],
    'inputs': {},
    'outputs': {
        'result': {
            'title': 'List of supported schemas',
            'description': 'List of supported schemas',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json'
            }
        }
    },
    'example': {
        'inputs': {}
    }
}
PROCESS_METADATA_IMPORT = {
    'version': __version__,
    'id': 'pygeometa-metadata-import',
    'title': {
        'en': 'pygeometa metadata import',
    },
    'description': {
        'en': 'Import metadata into MCF'
    },
    'keywords': ['pygeometa', 'metadata', 'import'],
    'links': [{
        'type': 'text/html',
        'rel': 'about',
        'title': 'information',
        'href': 'https://geopython.github.io/pygeometa/pygeoapi-plugin',
        'hreflang': 'en-US'
    }],
    'inputs': {
        'metadata': INPUT_METADATA,
        'schema': INPUT_SCHEMA
    },
    'outputs': {
        'result': {
            'title': 'Generated MCF',
            'description': 'Generated MCF',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json'
            }
        }
    },
    'example': {
        'inputs': {
            'metadata': '<gmd:MD_Metadata>...</gmd:MD_Metadata>',
            'schema': 'iso19139'
        }
    }
}

PROCESS_METADATA_VALIDATE = {
    'version': __version__,
    'id': 'pygeometa-metadata-validate',
    'title': {
        'en': 'pygeometa metadata control file (MCF) validation',
    },
    'description': {
        'en': 'Validate metadata from a pygeometa metadata control file (MCF)',
    },
    'keywords': ['pygeometa', 'metadata', 'validate'],
    'links': [{
        'type': 'text/html',
        'rel': 'about',
        'title': 'information',
        'href': 'https://geopython.github.io/pygeometa/pygeoapi-plugin',
        'hreflang': 'en-US'
    }],
    'inputs': {
        'mcf': INPUT_MCF
    },
    'outputs': {
        'result': {
            'title': 'Validate report',
            'description': 'Validate report',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json'
            }
        }
    },
    'example': {
        'inputs': {
            'mcf': {'mcf': {'version': '1.0'}}
        }
    }
}


PROCESS_METADATA_GENERATE = {
    'version': __version__,
    'id': 'pygeometa-metadata-generate',
    'title': {
        'en': 'pygeometa metadata generation',
    },
    'description': {
        'en': 'Generate metadata from a pygeometa metadata control file (MCF)'
    },
    'keywords': ['pygeometa', 'metadata', 'generate'],
    'links': [{
        'type': 'text/html',
        'rel': 'about',
        'title': 'information',
        'href': 'https://geopython.github.io/pygeometa/pygeoapi-plugin',
        'hreflang': 'en-US'
    }],
    'inputs': {
        'mcf': INPUT_MCF,
        'schema': INPUT_SCHEMA
    },
    'outputs': {
        'result': {
            'title': 'Generated metadata',
            'description': 'Generated metadata',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json'
            }
        }
    },
    'example': {
        'inputs': {
            'mcf': {'mcf': {'version': '1.0'}},
            'schema': 'oarec-record'
        }
    }
}


PROCESS_METADATA_TRANSFORM = {
    'version': __version__,
    'id': 'pygeometa-metadata-transform',
    'title': {
        'en': 'pygeometa metadata transformation',
    },
    'description': {
        'en': 'Transform metadata'
    },
    'keywords': ['pygeometa', 'metadata', 'transform'],
    'links': [{
        'type': 'text/html',
        'rel': 'about',
        'title': 'information',
        'href': 'https://geopython.github.io/pygeometa/pygeoapi-plugin',
        'hreflang': 'en-US'
    }],
    'inputs': {
        'metadata': INPUT_METADATA,
        'input-schema': INPUT_SCHEMA,
        'output-schema': INPUT_SCHEMA
    },
    'outputs': {
        'result': {
            'title': 'Transformed metadata',
            'description': 'Transformed metadata',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json'
            }
        }
    },
    'example': {
        'inputs': {
            'metadata': '<gmd:MD_Metadata>...</gmd:MD_Metadata>',
            'input-schema': 'iso19139',
            'output-schema': 'oarec-record'
        }
    }
}


class PygeometaMetadataSchemasProcessor(BaseProcessor):
    """pygeometa metadata schemas example"""

    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pygeometa.pygeoapi_plugin.PygeometaMetadataSchemasProcessor
        """

        super().__init__(processor_def, PROCESS_METADATA_SCHEMAS)

    def execute(self, data):

        mimetype = 'application/json'

        outputs = {
            'id': 'schemas',
            'value': get_supported_schemas(details=True)
        }

        return mimetype, outputs

    def __repr__(self):
        return '<PygeometaMetadataSchemasProcessor>'


class PygeometaMetadataImportProcessor(BaseProcessor):
    """pygeometa metadata import example"""

    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pygeometa.pygeoapi_plugin.PygeometaMetadataImportProcessor
        """

        super().__init__(processor_def, PROCESS_METADATA_IMPORT)

    def execute(self, data):

        response = None
        mimetype = 'application/json'
        metadata = data.get('metadata')
        schema = data.get('schema')

        if None in [metadata, schema]:
            msg = 'Missing metadata or schema'
            LOGGER.error(msg)
            raise ProcessorExecuteError(msg)

        try:
            LOGGER.info(f'Importing metadata into {schema}')
            schema_object = load_schema(schema)
            response = schema_object.import_(metadata)
        except NotImplementedError:
            msg = f'Import not supported for {schema}'
            response = msg
            raise ProcessorExecuteError(msg)
        except Exception as err:
            msg = f'Invalid metadata: {err}'
            response = msg

        outputs = {
            'id': 'import-report',
            'value': response
        }

        return mimetype, outputs

    def __repr__(self):
        return '<PygeometaMetadataImportProcessor>'


class PygeometaMetadataValidateProcessor(BaseProcessor):
    """pygeometa metadata validate example"""

    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pygeometa.pygeoapi_plugin.PygeometaMetadataValidateProcessor
        """

        super().__init__(processor_def, PROCESS_METADATA_VALIDATE)

    def execute(self, data):

        response = None
        mimetype = 'application/json'
        mcf = data.get('mcf')

        if mcf is None:
            msg = 'Missing input MCF'
            LOGGER.error(msg)
            raise ProcessorExecuteError(msg)

        try:
            LOGGER.debug('Validating MCF')
            instance = read_mcf(mcf)
            validate_mcf(instance)
            response = 'Valid MCF'
        except Exception as err:
            response = f'Invalid MCF: {err}'

        outputs = {
            'id': 'validate-report',
            'value': response
        }

        return mimetype, outputs

    def __repr__(self):
        return '<PygeometaMetadataValidateProcessor>'


class PygeometaMetadataGenerateProcessor(BaseProcessor):
    """pygeometa metadata generate example"""

    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pygeometa.pygeoapi_plugin.PygeometaMetadataGenerateProcessor
        """

        super().__init__(processor_def, PROCESS_METADATA_GENERATE)

    def execute(self, data):

        response = None
        mimetype = 'application/json'
        mcf = data.get('mcf')
        schema = data.get('schema')

        if None in [mcf, schema]:
            msg = 'Missing input MCF or schema'
            LOGGER.error(msg)
            raise ProcessorExecuteError(msg)

        try:
            LOGGER.debug('Generating metadata')
            instance = read_mcf(mcf)
            schema_object = load_schema(schema)

            if schema_object.outputformat == 'json':
                stringify = False
            else:
                mimetype = 'application/xml'
                stringify = True

            response = schema_object.write(instance, stringify=stringify)

        except Exception as err:
            response = f'Generation error: {err}'

        return mimetype, response

    def __repr__(self):
        return '<PygeometaMetadataGenerateProcessor>'


class PygeometaMetadataTransformProcessor(BaseProcessor):
    """pygeometa metadata transform example"""

    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pygeometa.pygeoapi_plugin.PygeometaMetadataGenerateProcessor
        """

        super().__init__(processor_def, PROCESS_METADATA_TRANSFORM)

    def execute(self, data):

        content = None
        response = None
        mimetype = 'application/json'
        metadata = data.get('metadata')
        input_schema = data.get('input-schema')
        output_schema = data.get('output-schema')

        if None in [metadata, input_schema, output_schema]:
            msg = 'Missing metadata or input-schema or output-schema'
            LOGGER.error(msg)
            raise ProcessorExecuteError(msg)

        try:
            LOGGER.info(f'Importing {metadata} into {input_schema}')
            schema_object_input = load_schema(input_schema)
            content = schema_object_input.import_(metadata)
        except NotImplementedError:
            msg = f'Import not supported for {input_schema}'
            LOGGER.error(msg)
            raise ProcessorExecuteError(msg)

        schema_object_output = load_schema(output_schema)

        if schema_object_output.outputformat == 'json':
            stringify = False
        else:
            mimetype = 'application/xml'
            stringify = True

        response = schema_object_output.write(content, stringify=stringify)

        return mimetype, response

    def __repr__(self):
        return '<PygeometaMetadataGenerateProcessor>'



# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import base64
from datetime import date, datetime, time
from decimal import Decimal
import logging
from pathlib import Path

LOGGER = logging.getLogger(__name__)

THISDIR = Path(__file__).resolve().parent


def json_serial(obj):
    """
    helper function to convert to JSON non-default
    types (source: https://stackoverflow.com/a/22238613)

    :param obj: `object` to be evaluated

    :returns: JSON non-default type to `str`
    """

    if isinstance(obj, (datetime, date, time)):
        return obj.isoformat()
    elif isinstance(obj, bytes):
        try:
            LOGGER.debug('Returning as UTF-8 decoded bytes')
            return obj.decode('utf-8')
        except UnicodeDecodeError:
            LOGGER.debug('Returning as base64 encoded JSON object')
            return base64.b64encode(obj)
    elif isinstance(obj, Decimal):
        return float(obj)

    msg = f'{obj} type {type(obj)} not serializable'
    LOGGER.error(msg)
    raise TypeError(msg)



# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import ast
import logging
import os

from lxml import etree
from owslib.iso import CI_OnlineResource, CI_ResponsibleParty, MD_Metadata

from pygeometa.schemas.base import BaseOutputSchema

LOGGER = logging.getLogger(__name__)
THISDIR = os.path.dirname(os.path.realpath(__file__))


class ISO19139OutputSchema(BaseOutputSchema):
    """ISO 19139 output schema"""

    def __init__(self):
        """
        Initialize object

        :returns: pygeometa.schemas.base.BaseOutputSchema
        """

        description = 'ISO 19115/19139'

        super().__init__('iso19139', description, 'xml', THISDIR)

    def import_(self, metadata: str) -> dict:
        """
        Import metadata into MCF

        :param metadata: string of metadata content

        :returns: `dict` of MCF content
        """

        mcf = {
            'mcf': {
                'version': '1.0',
            },
            'metadata': {},
            'identification': {},
            'contact': {},
            'distribution': {}
        }

        LOGGER.debug('Parsing ISO metadata')
        try:
            m = MD_Metadata(etree.fromstring(metadata))
        except ValueError:
            m = MD_Metadata(etree.fromstring(bytes(metadata, 'utf-8')))

        LOGGER.debug('Setting metadata')
        mcf['metadata']['identifier'] = m.identifier

        mcf['metadata']['hierarchylevel'] = m.hierarchy
        mcf['metadata']['datestamp'] = m.datestamp

        LOGGER.debug('Setting language')
        if m.language:
            mcf['metadata']['language'] = m.language
        elif m.languagecode:
            mcf['metadata']['language'] = m.languagecode

        identification = m.identification[0]

        LOGGER.debug('Setting identification')
        mcf['identification']['title'] = identification.title
        mcf['identification']['abstract'] = identification.abstract

        if identification.date:
            mcf['identification']['dates'] = {}
            for date_ in identification.date:
                mcf['identification']['dates'][date_.type] = date_.date

        if identification.keywords:
            mcf['identification']['keywords'] = {}
            for count, value in enumerate(identification.keywords):
                key = f'keywords-{count}'
                mcf['identification']['keywords'][key] = {
                    'keywords_type': value.type,
                    'keywords': [k.name for k in value.keywords]
                }
                if value.thesaurus is not None:
                    mcf['identification']['keywords'][key]['vocabulary'] = {
                        'name': value.thesaurus['title'],
                        'url': value.thesaurus['url']
                    }

        mcf['identification']['topiccategory'] = identification.topiccategory  # noqa

        mcf['identification']['extents'] = {
            'spatial': [{
                'bbox': [
                    ast.literal_eval(identification.extent.boundingBox.minx),
                    ast.literal_eval(identification.extent.boundingBox.miny),
                    ast.literal_eval(identification.extent.boundingBox.maxx),
                    ast.literal_eval(identification.extent.boundingBox.maxy)
                ]
            }],
            'temporal': []
        }

        temp_extent = {
            'begin': None,
            'end': None
        }

        if identification.temporalextent_start:
            temp_extent['begin'] = identification.temporalextent_start
        if identification.temporalextent_end:
            temp_extent['end'] = identification.temporalextent_end

        mcf['identification']['extents']['temporal'].append(temp_extent)

        if identification.accessconstraints:
            mcf['identification']['accessconstraints'] = identification.accessconstraints[0]  # noqa

        mcf['identification']['status'] = identification.status

        LOGGER.debug('Setting contacts')
#        for contact in m.get_all_contacts():
#            mcf['contact'].update(get_contact(contact))
        mcf['contact'].update(get_contact(m.contact[0]))

        LOGGER.debug('Setting distribution')
        if m.distribution:
            for count, value in enumerate(m.distribution.online):
                key = f'link-{count}'
                mcf['distribution'][key] = get_link(value)

        return mcf


def get_contact(contact: CI_ResponsibleParty) -> dict:
    """
    Generates an MCF contact from an OWSLib contact

    :param contact: OWSLib `CI_ResponsibleParty` object

    :returns: dict of MCF contact
    """

    mcf_contact = {contact.role: {}}

    cm_lookup = {
        'name': 'name',
        'organization': 'organization',
        'positionname': 'position',
        'phone': 'phone',
        'fax': 'fax',
        'address': 'address',
        'city': 'city',
        'administrativearea': 'region',
        'postalcode': 'postcode',
        'country': 'country',
        'email': 'email'
    }

    for key, value in cm_lookup.items():
        if getattr(contact, value) is not None:
            mcf_contact[contact.role][key] = getattr(contact, value)

    if hasattr(contact.onlineresource, 'url'):
        mcf_contact[contact.role]['url'] = contact.onlineresource.url

    return mcf_contact


def get_link(link: CI_OnlineResource) -> dict:
    """
    Generates an MCF link from an OWSLib distribution URL

    :param contact: OWSLib `CI_OnlineResource` object

    :returns: dict of MCF link
    """

    mcf_link = {
        'url': link.url,
        'type': link.protocol,
        'name': link.name,
        'description': link.description,
        'function': link.function
    }

    return mcf_link



# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2020 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import os

from pygeometa.schemas.base import BaseOutputSchema

THISDIR = os.path.dirname(os.path.realpath(__file__))


class WMOCMPOutputSchema(BaseOutputSchema):
    """WMO Core Metadata Profile output schema"""

    def __init__(self):
        """
        Initialize object

        :returns: pygeometa.schemas.base.BaseOutputSchema
        """

        description = 'WMO Core Metadata Profile (WCMP)'
        super().__init__('wmo-cmp', description, 'xml', THISDIR)



# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2020 Tom Kralidis, Paul van Genuchten
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import json
import os
from typing import Union

from pygeometa.helpers import json_serial
from pygeometa.schemas.base import BaseOutputSchema

THISDIR = os.path.dirname(os.path.realpath(__file__))


class DCATOutputSchema(BaseOutputSchema):
    """dcat output schema"""

    def __init__(self):
        """
        Initialize object

        :returns: pygeometa.schemas.base.BaseOutputSchema
        """

        description = 'DCAT'
        super().__init__('dcat', description, 'json', THISDIR)

    def write(self, mcf: dict, stringify: str = True) -> Union[dict, str]:
        """
        Write MCF to DCAT

        :param mcf: dict of MCF content model
        :param stringify: whether to return a string representation (default)
                          else native (dict, etree)

        :returns: `dict` or `str` of MCF as a DCAT representation
        """

        dcat = {
            "@context": {
                # namespaces
                "adms": "http://www.w3.org/ns/adms#",
                "dcat": "http://www.w3.org/ns/dcat#",
                "dct": "http://purl.org/dc/terms/",
                "foaf": "http://xmlns.com/foaf/0.1/",
                "gsp": "http://www.opengis.net/ont/geosparql#",
                "locn": "http://www.w3.org/ns/locn#",
                "owl": "http://www.w3.org/2002/07/owl#",
                "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "schema": "http://schema.org/",
                "skos": "http://www.w3.org/2004/02/skos/core#",
                "time": "http://www.w3.org/2006/time",
                "vcard": "http://www.w3.org/2006/vcard/ns#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                # mcf-property to dcat mappings
                "topiccategory": "dcat:theme",
                "language": "dct:language",
                # todo: support any range of languages from source
                "title": "dct:title",
                "title_en": {"@id": "dct:title", "@language": "en"},
                "title_fr": {"@id": "dct:title", "@language": "fr"},
                "abstract_en": {"@id": "dct:description", "@language": "en"},
                "abstract_fr": {"@id": "dct:description", "@language": "fr"},
                "distribution": "dcat:distribution",
                "url": "dcat:accessURL",
                "name": "dct:title",
                "name_en": {"@id": "dct:title", "@language": "en"},
                "name_fr": {"@id": "dct:title", "@language": "fr"},
                "description": "dct:description",
                "description_en": {"@id": "dct:description",
                                   "@language": "en"},
                "description_fr": {"@id": "dct:description",
                                   "@language": "fr"},
                "keywords": "dct:keyword",
                "keywords_en": {"@id": "dct:keyword", "@language": "en"},
                "keywords_fr": {"@id": "dct:keyword", "@language": "fr"},
                "dataseturi": {"@type": "@id", "@id": "@id"},
                "contact": "dcat:contactPoint",
                "spatial": "dct:spatial",
                "temporal": "dct:temporal",
                "creation": "dct:issued",
                "modified": "dct:modified",
                "maintenancefrequency": "dct:accrualPeriodicity",
                "type": "dcat:mediaType",
                "size": "dcat:byteSize",
                "status": "adms:status",
                "organization": "vcard:hasOrganizationName",
                "individualname": "vcard:fn",
                "phone": "vcard:hasTelephone",
                "address": "vcard:street-address",
                "city": "vcard:locality",
                "postalcode": "vcard:postal-code",
                "country": "vcard:country-name",
                "email": "vcard:hasEmail",
                "accessconstraints": "dct:accessRights",
                "rights": "dct:rights",
                "rights_en": {"@id": "dct:rights", "@language": "en"},
                "rights_fr": {"@id": "dct:rights", "@language": "fr"},
                "bbox": "dcat:bbox",
                "begin": "dcat:startDate",
                "end": "dcat:endDate"
            },
            "@type": "dcat:Dataset",
            "keywords": [],
            "keywords_en": [],
            "keywords_fr": [],
            "distribution": [],
            "contact": []
        }

        # prepare mcf for json-ld
        for key, value in mcf.items():
            # do nothing for these items (yet)
            if (key in ['mcf', 'content_info', 'acquisition']):
                None
            # unnest these items
            elif (key in ['metadata', 'identification']):
                for k, v in value.items():
                    if (k == 'extents'):
                        for k1, v1 in v.items():
                            # assign dct:location type
                            if (k1 == 'spatial'):
                                dcat["spatial"] = []
                                for k2 in v1:
                                    k2['@type'] = 'dct:Location'
                                    dcat["spatial"].append(k2)
                            # assign dct:PeriodOftime type
                            elif (k1 == 'temporal'):
                                dcat['temporal'] = []
                                for k3 in v1:
                                    k3['@type'] = 'dct:PeriodOfTime'
                                    dcat["temporal"].append(k3)
                    # unnest keywords
                    elif (k == 'keywords'):
                        for k4, v4 in v.items():
                            for k5, v5 in v4.items():
                                if (k5 != 'keywords_type'):
                                    for kw in v5:
                                        # assumes a key for language exists
                                        if k5 in dcat:
                                            dcat[k5].append(kw)
                    elif (k in ['identifier']):
                        # mint a url from identifier if non exists on mcf
                        if (not mcf['metadata']['dataseturi']):
                            dcat['dataseturi'] = 'http://example.com/#' + \
                                str(v)
                    elif (k in ['dates']):
                        for k6, v6 in v.items():
                            dcat[k6] = v6
                    else:
                        dcat[k] = v
            # transform set of keys to array
            elif (key in ['distributor', 'contact']):
                for k, v in value.items():
                    # add id (if url exists)
                    if (not isinstance(v, str) and v['url']):
                        v['@id'] = v['url']
                    # add type
                    if (key == 'distribution'):
                        v['@type'] = 'dcat:Distribution'
                    else:
                        v['@type'] = 'vcard:Organization'
                    dcat[key].append(v)
            # other cases: root properties
            else:
                dcat[key] = value

        if stringify:
            return json.dumps(dcat, default=json_serial, indent=4)

        return dcat



# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2016 Government of Canada
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
import sys

import click

ARGUMENT_MCF = click.argument('mcf')
ARGUMENT_METADATA_FILE = click.argument('metadata-file', type=click.File())

OPTION_OUTPUT = click.option(
    '--output',
    '-o',
    type=click.File('w', encoding='utf-8'),
    help='Name of output file')


def OPTION_VERBOSITY(f):
    logging_options = ['ERROR', 'WARNING', 'INFO', 'DEBUG']

    def callback(ctx, param, value):
        if value is not None:
            logging.basicConfig(stream=sys.stdout,
                                level=getattr(logging, value))
        return True

    return click.option('--verbosity', '-v',
                        type=click.Choice(logging_options),
                        help='Verbosity',
                        callback=callback)(f)


def cli_callbacks(f):
    f = OPTION_VERBOSITY(f)
    return f



# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2016 Government of Canada
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from pathlib import Path
from setuptools import Command, find_packages, setup
import re
import sys


class PyTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess
        errno = subprocess.call([sys.executable, 'tests/run_tests.py'])
        raise SystemExit(errno)


def read(filename) -> str:
    """read file contents"""

    fullpath = Path(__file__).resolve().parent / filename

    with fullpath.open() as fh:
        contents = fh.read().strip()

    return contents


def get_package_version():
    """get version from top-level package init"""

    version_file = read('pygeometa/__init__.py')
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


LONG_DESCRIPTION = read('README.md')

DESCRIPTION = 'pygeometa is a Python package to generate metadata for geospatial datasets'  # noqa

MANIFEST = Path('MANIFEST')

if MANIFEST.exists():
    MANIFEST.unlink()

setup(
    name='pygeometa',
    version=get_package_version(),
    description=DESCRIPTION.strip(),
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    license='MIT',
    platforms='all',
    keywords=' '.join([
        'geospatial',
        'metadata',
        'catalogue',
        'discovery'
    ]),
    author='Meteorological Service of Canada',
    author_email='tom.kralidis@canada.ca',
    maintainer='Meteorological Service of Canada',
    maintainer_email='tom.kralidis@canada.ca',
    url='https://geopython.github.io/pygeometa',
    install_requires=read('requirements.txt').splitlines(),
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pygeometa=pygeometa:cli'
        ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: GIS'
    ],
    cmdclass={'test': PyTest},
    test_suite='tests.run_tests'
)



# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import importlib
import logging
import os

from pygeometa.schemas.base import BaseOutputSchema

LOGGER = logging.getLogger(__name__)
THISDIR = os.path.dirname(os.path.realpath(__file__))

SCHEMAS = {
    'iso19139': 'pygeometa.schemas.iso19139.ISO19139OutputSchema',
    'iso19139-2': 'pygeometa.schemas.iso19139_2.ISO19139_2OutputSchema',
    'iso19139-hnap': 'pygeometa.schemas.iso19139_hnap.ISO19139HNAPOutputSchema',  # noqa
    'oarec-record': 'pygeometa.schemas.ogcapi_records.OGCAPIRecordOutputSchema',  # noqa
    'stac-item': 'pygeometa.schemas.stac.STACItemOutputSchema',
    'dcat': 'pygeometa.schemas.dcat.DCATOutputSchema',
    'wmo-cmp': 'pygeometa.schemas.wmo_cmp.WMOCMPOutputSchema',
    'wmo-wcmp2': 'pygeometa.schemas.wmo_wcmp2.WMOWCMP2OutputSchema',
    'wmo-wigos': 'pygeometa.schemas.wmo_wigos.WMOWIGOSOutputSchema'
}


def get_supported_schemas(details: bool = False) -> list:
    """
    Get supported schemas

    :param details: provide read/write details

    :returns: list of supported schemas
    """

    def has_mode(plugin: BaseOutputSchema, mode: str) -> bool:
        enabled = False

        try:
            _ = getattr(plugin, mode)('test')
        except NotImplementedError:
            pass
        except Exception:
            enabled = True

        return enabled

    schema_matrix = []

    LOGGER.debug('Generating list of supported schemas')

    if not details:
        return SCHEMAS.keys()

    for key in SCHEMAS.keys():
        schema = load_schema(key)
        can_read = has_mode(schema, 'import_')
        can_write = has_mode(schema, 'write')

        schema_matrix.append({
            'id': key,
            'description': schema.description,
            'read': can_read,
            'write': can_write
        })

    return schema_matrix


def load_schema(schema_name: str) -> BaseOutputSchema:
    """
    loads schema plugin by name

    :param schema_name: shortname of schema

    :returns: plugin object
    """

    LOGGER.debug(f'Schemas: {SCHEMAS.keys()}')

    if schema_name not in SCHEMAS.keys():
        msg = f'Schema {schema_name} not found'
        LOGGER.exception(msg)
        raise InvalidSchemaError(msg)

    name = SCHEMAS[schema_name]

    if '.' in name:  # dotted path
        packagename, classname = name.rsplit('.', 1)
    else:
        raise InvalidSchemaError(f'Schema path {name} not found')

    LOGGER.debug(f'package name: {packagename}')
    LOGGER.debug(f'class name: {classname}')

    module = importlib.import_module(packagename)
    class_ = getattr(module, classname)

    return class_()


class InvalidSchemaError(Exception):
    """Invalid plugin"""
    pass



# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2024 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from datetime import date, datetime
import json
import logging
import os
from typing import Union

from pygeometa.core import get_charstring
from pygeometa.helpers import json_serial
from pygeometa.schemas.base import BaseOutputSchema

THISDIR = os.path.dirname(os.path.realpath(__file__))

LOGGER = logging.getLogger(__name__)


class OGCAPIRecordOutputSchema(BaseOutputSchema):
    """OGC API - Records - Part 1: Core record schema"""

    def __init__(self):
        """
        Initialize object

        :returns: pygeometa.schemas.base.BaseOutputSchema
        """

        description = 'OGC API - Records - Part 1: Record Core'

        super().__init__('oarec-record', description, 'json', THISDIR)

    def write(self, mcf: dict, stringify: str = True) -> Union[dict, str]:
        """
        Write outputschema to JSON string buffer

        :param mcf: dict of MCF content model
        :param stringify: whether to return a string representation (default)
                          else native (dict, etree)


        :returns: `dict` or `str` of MCF as an OARec record representation
        """

        self.lang1 = mcf['metadata'].get('language')
        self.lang2 = mcf['metadata'].get('language_alternate')

        minx, miny, maxx, maxy = (mcf['identification']['extents']
                                  ['spatial'][0]['bbox'])

        title = get_charstring(mcf['identification'].get('title'),
                               self.lang1, self.lang2)

        description = get_charstring(mcf['identification'].get('abstract'),
                                     self.lang1, self.lang2)

        LOGGER.debug('Generating baseline record')
        record = {
            'id': mcf['metadata']['identifier'],
            'conformsTo': [
                'http://www.opengis.net/spec/ogcapi-records-1/1.0/req/record-core',  # noqa
            ],
            'type': 'Feature',
            'geometry': {
                'type': 'Polygon',
                'coordinates': [[
                    [minx, miny],
                    [minx, maxy],
                    [maxx, maxy],
                    [maxx, miny],
                    [minx, miny]
                ]]
            },
            'properties': {
                'identifier': mcf['metadata']['identifier'],
                'title': title[0],
                'description': description[0],
                'themes': [],
                'language': self.lang1,
                'type': mcf['metadata']['hierarchylevel'],
            },
            'links': []
        }

        LOGGER.debug('Checking for temporal')
        try:
            begin = mcf['identification']['extents']['temporal'][0]['begin']
            end = mcf['identification']['extents']['temporal'][0].get('end')

            if begin in ['now', 'None', None]:
                begin = '..'
            else:
                begin = str(begin)

            if end in ['now', 'None', None]:
                end = '..'
            else:
                end = str(end)

            if [begin, end] == [None, None]:
                record['time'] = None
            else:
                record['time'] = {
                    'interval': [begin, end]
                }

            if 'resolution' in  mcf['identification']['extents']['temporal'][0]:  # noqa
                record['time']['resolution'] =  mcf['identification']['extents']['temporal'][0]['resolution']  # noqa

        except (IndexError, KeyError):
            record['time'] = None

        LOGGER.debug('Checking for dates')

        if 'dates' in mcf['identification']:
            if 'creation' in mcf['identification']['dates']:
                record['properties']['created'] = self.generate_date(mcf['identification']['dates']['creation'])  # noqa

            if 'revision' in mcf['identification']['dates']:
                record['properties']['updated'] = self.generate_date(mcf['identification']['dates']['revision'])  # noqa

        rights = get_charstring(mcf['identification'].get('rights'),
                                self.lang1, self.lang2)

        if rights != [None, None]:
            record['properties']['rights'] = rights[0]

        formats = []
        for v in mcf['distribution'].values():
            format_ = get_charstring(v.get('format'), self.lang1, self.lang2)
            if format_[0] is not None:
                formats.append(format_)

        LOGGER.debug('Checking for formats')
        if formats:
            record['properties']['formats'] = list(
                set([f[0] for f in formats]))

        LOGGER.debug('Checking for contacts')
        record['properties']['contacts'] = self.generate_contacts(
            mcf['contact'])

        all_keywords = []

        LOGGER.debug('Checking for keywords')
        for key, value in mcf['identification']['keywords'].items():
            theme = {'concepts': []}
            scheme = None

            keywords = get_charstring(value.get('keywords'), self.lang1,
                                      self.lang2)

            if 'vocabulary' in value:
                if 'url' in value['vocabulary']:
                    scheme = value['vocabulary']['url']
                elif 'name' in value['vocabulary']:
                    scheme = value['vocabulary']['name']

            if scheme is None:
                LOGGER.debug('Keywords found without vocabulary')
                LOGGER.debug('Aggregating as bare keywords')
                all_keywords.extend(keywords[0])
            else:
                LOGGER.debug('Adding as theme/concepts')
                for kw in keywords[0]:
                    theme['concepts'].append({'id': kw})

                theme['scheme'] = scheme

                record['properties']['themes'].append(theme)

        if all_keywords:
            record['properties']['keywords'] = all_keywords

        LOGGER.debug('Checking for licensing')
        if mcf['identification'].get('license') is not None:
            license = mcf['identification']['license']

            if 'url' in license:
                LOGGER.debug('Encoding license as link')
                license_link = {
                    'rel': 'license',
                    'type': 'text/html',
                    'title': license.get('name', 'license for this resource'),
                    'url': license['url']
                }
                record['links'].append(self.generate_link(license_link))
            else:
                LOGGER.debug('Encoding license as property')
                record['properties']['license'] = license['name']

        LOGGER.debug('Checking for distribution')
        for value in mcf['distribution'].values():
            record['links'].append(self.generate_link(value))

        if stringify:
            return json.dumps(record, default=json_serial, indent=4)

        return record

    def generate_party(self, contact: dict,
                       lang1: str, lang2: str, roles: list) -> dict:
        """
        generate party construct from MCF contact

        :param contact: dict of MCF contact
        :param self.lang1: primary language
        :param self.lang2: alternate language
        :param roles: roles of contact

        :returns: MCF contact as a party representation
        """

        organization_name = get_charstring(contact.get('organization'),
                                           self.lang1, self.lang2)

        individual_name = get_charstring(contact.get('individualname'),
                                         self.lang1, self.lang2)

        position_name = get_charstring(contact.get('positionname'),
                                       self.lang1, self.lang2)

        hours_of_service = get_charstring(contact.get('hoursofservice'),
                                          self.lang1, self.lang2)

        contact_instructions = get_charstring(
            contact.get('contactinstructions'), self.lang1, self.lang2)

        address = get_charstring(contact.get('address'),
                                 self.lang1, self.lang2)

        city = get_charstring(contact.get('city'), self.lang1, self.lang2)

        administrative_area = get_charstring(contact.get('administrativearea'),
                                             self.lang1, self.lang2)

        postalcode = get_charstring(contact.get('postalcode'),
                                    self.lang1, self.lang2)

        country = get_charstring(contact.get('country'),
                                 self.lang1, self.lang2)

        rp = {
            'addresses': [{}],
            'roles': []
        }

        if organization_name[0] is not None:
            rp['organization'] = organization_name[0]
        if individual_name[0] is not None:
            rp['name'] = individual_name[0]
        if position_name[0] is not None:
            rp['position'] = position_name[0]
        if hours_of_service[0] is not None:
            rp['hoursOfService'] = hours_of_service[0]
        if contact_instructions[0] is not None:
            rp['contactInstructions'] = contact_instructions[0]

        if address[0] is not None:
            rp['addresses'][0]['deliveryPoint'] = [address[0]]
        if city[0] is not None:
            rp['addresses'][0]['city'] = city[0]
        if administrative_area[0] is not None:
            rp['addresses'][0]['administrativeArea'] = administrative_area[0]
        if postalcode[0] is not None:
            rp['addresses'][0]['postalCode'] = postalcode[0]
        if country[0] is not None:
            rp['addresses'][0]['country'] = country[0]

        if contact.get('phone') is not None:
            LOGGER.debug('Formatting phone number')
            phone = contact['phone']
            phone = phone.replace('-', '').replace('(', '').replace(')', '')
            phone = phone.replace('+0', '+').replace(' ', '')

            rp['phones'] = [{'value': phone}]

        if contact.get('email') is not None:
            rp['emails'] = [{'value': contact.get('email')}]

        if rp['addresses'][0] == {}:
            rp.pop('addresses')

        for r in set(roles):
            rp['roles'].append(r)

        if 'url' in contact:
            rp['links'] = [{
                'rel': 'canonical',
                'type': 'text/html',
                'href': contact['url']
            }]

        return rp

    def generate_contacts(self, contact: dict) -> list:
        """
        Generates 1..n contacts, streamlining identical
        contacts with multiple roles

        :param contact: `dict` of contacts

        :returns: `list` of contacts
        """

        contacts = []
        contacts2 = []

        for key, value in contact.items():
            if contacts:
                for c in contacts:
                    if value == c['contact']:
                        LOGGER.debug('Found matching contact; adding role')
                        c['roles'].append(key)
                    else:
                        LOGGER.debug('Adding contact')
                        contacts.append({
                            'contact': value,
                            'roles': [key]
                        })
            else:
                contacts.append({
                    'contact': value,
                    'roles': [key]
                })

        LOGGER.debug(f'Contacts: {contacts}')
        for c in contacts:
            contacts2.append(self.generate_party(c['contact'], self.lang1,
                             self.lang2, c['roles']))

        return contacts2

    def generate_link(self, distribution: dict) -> dict:
        """
        Generates OARec link object from MCF distribution object

        :param distribution: `dict` of MCF distribution

        :returns: OARec link object
        """

        title = get_charstring(distribution.get('title'),
                               self.lang1, self.lang2)

        name = get_charstring(distribution.get('name'), self.lang1, self.lang2)

        link = {
            'href': distribution['url'],
            'type': distribution['type']
        }

        reltype = distribution.get('rel') or distribution.get('function')
        if reltype is not None:
            link['rel'] = reltype

        if title != [None, None]:
            link['title'] = title[0]
        elif name != [None, None]:
            link['title'] = name[0]

        if all(x in distribution['url'] for x in ['{', '}']):
            link['templated'] = True

        if 'channel' in distribution:
            link['channel'] = distribution['channel']

        return link

    def generate_date(self, date_value: str) -> str:
        """
        Helper function to derive RFC3339 date from MCF date type

        :param date_value: `str` of date value

        :returns: `str` of date-time value
        """

        value = None

        if isinstance(date_value, str) and date_value != 'None':
            if len(date_value) == 10:  # YYYY-MM-DD
                format_ = '%Y-%m-%d'
            elif len(date_value) == 7:  # YYYY-MM
                format_ = '%Y-%m'
            elif len(date_value) == 4:  # YYYY
                format_ = '%Y'

            LOGGER.debug('date type found; expanding to date-time')
            value = datetime.strptime(date_value, format_).strftime('%Y-%m-%dT%H:%M:%SZ')  # noqa

        elif isinstance(date_value, int) and len(str(date_value)) == 4:
            date_value2 = str(date_value)
            LOGGER.debug('date type found; expanding to date-time')
            format_ = '%Y'
            value = datetime.strptime(date_value2, format_).strftime('%Y-%m-%dT%H:%M:%SZ')  # noqa

        elif isinstance(date_value, (date, datetime)):
            value = date_value.strftime('%Y-%m-%dT%H:%M:%SZ')

        elif date_value in [None, 'None']:
            value = datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')

        else:
            msg = f'Unknown date string: {date_value}'
            raise RuntimeError(msg)

        return value



# =================================================================
#
# Terms and Conditions of Use
#
# Unless otherwise noted, computer program source code of this
# distribution # is covered under Crown Copyright, Government of
# Canada, and is distributed under the MIT License.
#
# The Canada wordmark and related graphics associated with this
# distribution are protected under trademark law and copyright law.
# No permission is granted to use them outside the parameters of
# the Government of Canada's corporate identity program. For
# more information, see
# http://www.tbs-sct.gc.ca/fip-pcim/index-eng.asp
#
# Copyright title to all 3rd party software distributed with this
# software is held by the respective copyright holders as noted in
# those files. Users are asked to read the 3rd Party Licenses
# referenced with those assets.
#
# Copyright (c) 2024 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from datetime import datetime
import json
import logging
import os
from typing import Union

from pygeometa.helpers import json_serial
from pygeometa.schemas.ogcapi_records import OGCAPIRecordOutputSchema

THISDIR = os.path.dirname(os.path.realpath(__file__))

LOGGER = logging.getLogger(__name__)


class WMOWCMP2OutputSchema(OGCAPIRecordOutputSchema):
    """OGC API - Records - Part 1: Core record schema"""

    def __init__(self):
        """
        Initialize object

        :returns: pygeometa.schemas.base.BaseOutputSchema
        """

        description = 'WMO Core Metadata Profile 2 (WCMP2)'

        super().__init__()

        self.description = description

    def write(self, mcf: dict, stringify: str = True) -> Union[dict, str]:
        """
        Write outputschema to JSON string buffer

        :param mcf: dict of MCF content model
        :param stringify: whether to return a string representation (default)
                          else native (dict, etree)


        :returns: `dict` or `str` of MCF as an OARec record representation
        """

        record = super().write(mcf, stringify=False)

        LOGGER.debug('Setting WCMP2 conformance')
        record['conformsTo'] = ['http://wis.wmo.int/spec/wcmp/2/conf/core']

        if 'edition' in mcf['identification']:
            record['properties']['version'] = mcf['identification']['edition']

        LOGGER.debug('Setting WCMP2 distribution links')
        record['links'] = []
        for key, value in mcf['distribution'].items():
            link = self.generate_link(value)

            record['links'].append(link)

        if mcf['metadata'].get('hierarchylevel') == 'dataset':
            try:
                record['properties']['wmo:dataPolicy'] = mcf['identification']['wmo_data_policy']  # noqa
            except KeyError:
                LOGGER.warning('Missing wmo:dataPolicy')

        if record['properties'].get('created') is None:
            record['properties']['created'] = datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')  # noqa

        if stringify:
            return json.dumps(record, default=json_serial, indent=4)
        else:
            return record
